/*

MLIST.C

Description:
Functions to handle the playlist.
Note: ironically, the tcl/tk interface will replace these functions
to interact with the playlist held in the frontend, not with mikmod.


Peter Amstutz <amstpi@freenet.tlh.fl.us>


HISTORY 
======

v1.00 (20/01/97) - first version

*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "mikmod.h"

void PL_InitList(PLAYLIST *pl)
{
	pl->numused=0;
	pl->current=-1;
	pl->file=NULL;
	pl->file_in_arc=NULL;
}

void PL_ClearList(PLAYLIST *pl)
{
	int i;
	for(i=0;i<pl->numused;i++)
	{
		free(pl->file[i]);
		pl->file=NULL;
		free(pl->file_in_arc[i]);
		pl->file_in_arc[i]=NULL;
	}
	if(pl->file!=NULL) {
		free(pl->file);
		pl->file=NULL;
	}
	if(pl->file_in_arc!=NULL) {
		free(pl->file_in_arc);
		pl->file_in_arc=NULL;
	}
	pl->current=0;
	pl->numused=0;
}

int PL_GetCurrentFile(PLAYLIST *pl, char *retfile, char *retarc)
{
	if(pl->numused==0) return 1;
	strcpy(retfile, pl->file[pl->current]);
	strcpy(retarc, pl->file_in_arc[pl->current]);
	if(pl->current==(pl->numused-1)) return 1;
	else return 0;
}

int PL_GetNext(PLAYLIST *pl, char *retfile, char *retarc)
{
	int ret=0;

	if(pl->numused==0) return 1;
	pl->current++;
	if(pl->current==pl->numused) {pl->current=pl->numused-1; ret=1;}

	if(retfile!=NULL) strcpy(retfile, pl->file[pl->current]);
	if(pl->file_in_arc[pl->current]!=NULL && retarc!=NULL) strcpy(retarc, pl->file_in_arc[pl->current]);

	return ret;
}

int PL_GetPrev(PLAYLIST *pl, char *retfile, char *retarc)
{
	if(pl->numused==0) return 1;
	pl->current--;
	if(pl->current<-1) pl->current=-1;

	if(retfile!=NULL) strcpy(retfile, pl->file[pl->current]);
	if(pl->file_in_arc[pl->current]!=NULL && retarc!=NULL) strcpy(retarc, pl->file_in_arc[pl->current]);

	return 0;
}

int PL_GetRandom(PLAYLIST *pl, char *retfile, char *retarc)
{
	if(pl->numused==0) return 1;
	srandom(time(NULL));
	pl->current=random() % pl->numused;
	strcpy(retfile,pl->file[pl->current]);
	strcpy(retfile,pl->file_in_arc[pl->current]);
	return 0;
}

int PL_DelCurrent(PLAYLIST *pl)
{
	int i;
	if(pl->numused==0) return 1;
	free(pl->file[pl->current]);
	free(pl->file_in_arc[pl->current]);

	for(i=pl->current;i<pl->numused;i++)
	{
		pl->file[i]=pl->file[i+1];
		pl->file_in_arc[i]=pl->file_in_arc[i+1];
	}
	pl->numused--;
	pl->current--;
	pl->file=realloc(pl->file,pl->numused*sizeof(char *));
	pl->file_in_arc=realloc(pl->file_in_arc,pl->numused*sizeof(char *));
	return 0;
}

void PL_Add(PLAYLIST *pl, char *file, char *arc)
{
	pl->numused++;
	pl->file=realloc(pl->file,pl->numused*sizeof(char *));
	pl->file_in_arc=realloc(pl->file_in_arc,pl->numused*sizeof(char *));

	pl->file[pl->numused-1]=malloc(strlen(file)+1);
	strcpy(pl->file[pl->numused-1],file);
	if(arc!=NULL) {
		pl->file_in_arc[pl->numused-1]=malloc(strlen(arc)+1);
		strcpy(pl->file_in_arc[pl->numused-1],arc);
	}
}

void PL_Load(PLAYLIST *pl, char *filename)
{
	FILE *f;
	char file[255];
	char *arc;

	if((f=fopen(filename,"r"))==NULL) {
		perror("Error loading playlist\n");
		return;
	}

	while(!feof(f))
	{
		fgets(file,255,f);
		if(file[strlen(file)-1]=='\n') file[strlen(file)-1]=0;
		arc=strchr(file,'|');
		if(arc==NULL)
		{
			PL_Add(pl, file, NULL);
		}
		else 
		{
			*arc=0;
			arc++;
			PL_Add(pl, file, arc);
		}
	}
	fclose(f);
}

void PL_Save(PLAYLIST *pl, char *savefile)
{

	FILE *f;
	int i;

	if((f=fopen(savefile,"r"))==NULL) {
		perror("Error saving playlist\n");
		return;
	}

	for(i=0;i<pl->numused;i++)
	{
		fprintf(f,"%s",pl->file[i]);
		if(pl->file_in_arc!=NULL) fprintf(f,"|%s",pl->file_in_arc[i]);
		fprintf(f,"\n");
	}
	fclose(f);
}
